import datetime as dt
from datetime import datetime, timedelta
from unittest.mock import patch

from django.utils.timezone import now, utc
from esi.errors import TokenError
from esi.models import Token
from eveuniverse.models import EveSolarSystem

from app_utils.testing import NoSocketsTestCase

from structures.models import Owner, OwnerCharacter
from structures.tests.testdata.factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
    EveSovereigntyMapFactory,
    OwnerCharacterFactory,
    OwnerFactory,
    UserMainBasicFactory,
    UserMainDefaultOwnerFactory,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.owners"


class TestOwner(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.alliance = EveAllianceInfoFactory(
            alliance_name="Wayne Enterprises", alliance_ticker="WYE"
        )
        corporation = EveCorporationInfoFactory(
            corporation_name="Wayne Technologies", alliance=cls.alliance
        )
        character = EveCharacterFactory(corporation=corporation)
        cls.user = UserMainDefaultOwnerFactory(main_character__character=character)
        cls.owner = OwnerFactory(user=cls.user)

    def test_str(self):
        # when
        result = str(self.owner)
        # then
        self.assertEqual(result, "Wayne Technologies")

    def test_repr(self):
        # when
        result = repr(self.owner)
        # then
        self.assertEqual(
            result, f"Owner(pk={self.owner.pk}, corporation='Wayne Technologies')"
        )

    @patch(MODULE_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    def test_is_structure_sync_fresh(self):
        # no errors and recent sync
        self.owner.structures_last_update_at = now()
        self.assertTrue(self.owner.is_structure_sync_fresh)

        # no errors and sync within grace period
        self.owner.structures_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_structure_sync_fresh)

        # no error, but no sync within grace period
        self.owner.structures_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_structure_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_is_notification_sync_fresh(self):
        # no errors and recent sync
        self.owner.notifications_last_update_at = now()
        self.assertTrue(self.owner.is_notification_sync_fresh)

        # no errors and sync within grace period
        self.owner.notifications_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_notification_sync_fresh)

        # no error, but no sync within grace period
        self.owner.notifications_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_notification_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_is_forwarding_sync_fresh(self):
        # no errors and recent sync
        self.owner.forwarding_last_update_at = now()
        self.assertTrue(self.owner.is_forwarding_sync_fresh)

        # no errors and sync within grace period
        self.owner.forwarding_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_forwarding_sync_fresh)

        # no error, but no sync within grace period
        self.owner.forwarding_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_forwarding_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    def test_is_assets_sync_fresh(self):
        # no errors and recent sync
        self.owner.assets_last_update_at = now()
        self.assertTrue(self.owner.is_assets_sync_fresh)

        # no errors and sync within grace period
        self.owner.assets_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_assets_sync_fresh)

        # no error, but no sync within grace period
        self.owner.assets_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_assets_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_is_all_syncs_ok(self):
        self.owner.structures_last_update_at = now()
        self.owner.notifications_last_update_at = now()
        self.owner.forwarding_last_update_at = now()
        self.owner.assets_last_update_at = now()
        self.assertTrue(self.owner.are_all_syncs_ok)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    def test_get_esi_scopes_pocos_off(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    def test_get_esi_scopes_pocos_on(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-planets.read_customs_offices.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    def test_get_esi_scopes_starbases_on(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_starbases.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    def test_get_esi_scopes_starbases_and_custom_offices(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_starbases.v1",
                "esi-planets.read_customs_offices.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    def test_should_ensure_only_one_owner_in_same_alliance_is_main(self):
        # given
        self.owner.is_alliance_main = True
        self.owner.save()
        corporation = EveCorporationInfoFactory(alliance=self.alliance)
        owner = OwnerFactory(corporation=corporation)
        # when
        owner.is_alliance_main = True
        owner.save()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_alliance_main)
        self.owner.refresh_from_db()
        self.assertFalse(self.owner.is_alliance_main)

    def test_should_ensure_is_alliance_main_is_set_for_update_fields_1(self):
        # given
        corporation = EveCorporationInfoFactory(alliance=self.alliance)
        owner = OwnerFactory(corporation=corporation, is_alliance_main=True)
        # when
        owner.save(update_fields={"is_active"})
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_alliance_main)

    def test_should_ensure_is_alliance_main_is_set_for_update_fields_2(self):
        # given
        corporation = EveCorporationInfoFactory(alliance=self.alliance)
        owner = OwnerFactory(corporation=corporation, is_alliance_main=True)
        # when
        owner.save(update_fields=["is_active"])
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_alliance_main)

    def test_should_allow_mains_from_other_alliances(self):
        # given
        self.owner.is_alliance_main = True
        self.owner.save()
        owner = OwnerFactory()
        # when
        owner.is_alliance_main = True
        owner.save()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_alliance_main)
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_alliance_main)

    def test_should_allow_other_corporations_to_be_main(self):
        # given
        self.owner.is_alliance_main = True
        self.owner.save()
        owner_2 = OwnerFactory(is_alliance_main=True)
        owner_3 = OwnerFactory()
        # when
        owner_3.is_alliance_main = True
        owner_3.save()
        # then
        owner_3.refresh_from_db()
        self.assertTrue(owner_3.is_alliance_main)
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_alliance_main)
        owner_2.refresh_from_db()
        self.assertTrue(owner_2.is_alliance_main)


class TestOwnerHasSov(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()
        EveSovereigntyMapFactory(
            corporation=cls.owner.corporation, eve_solar_system_name="1-PGSG"
        )

    def test_should_return_true_when_owner_has_sov(self):
        # given
        system = EveSolarSystem.objects.get(name="1-PGSG")
        # when/then
        self.assertTrue(self.owner.has_sov(system))

    def test_should_return_false_when_owner_has_no_sov(self):
        # given
        system = EveSolarSystem.objects.get(name="A-C5TC")
        # when/then
        self.assertFalse(self.owner.has_sov(system))

    def test_should_return_false_when_owner_is_outside_nullsec(self):
        # given
        system = EveSolarSystem.objects.get(name="Amamake")
        # when/then
        self.assertFalse(self.owner.has_sov(system))


@patch(MODULE_PATH + ".notify")
@patch(MODULE_PATH + ".notify_admins")
class TestOwnerFetchToken(NoSocketsTestCase):
    def test_should_return_correct_token(self, mock_notify_admins, mock_notify):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        # when
        token = owner.fetch_token()
        # then
        self.assertIsInstance(token, Token)
        self.assertEqual(token.user, user)
        self.assertEqual(token.character_id, character.character_id)
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            set(token.scopes.values_list("name", flat=True)),
        )
        self.assertFalse(mock_notify_admins.called)
        self.assertFalse(mock_notify.called)
        self.assertEqual(owner.characters.count(), 1)

    def test_raise_error_when_no_sync_char_defined(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = OwnerFactory(characters=False)
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertFalse(mock_notify_admins.called)
        self.assertFalse(mock_notify.called)

    def test_raise_error_when_user_has_no_permission_and_delete_character(
        self, mock_notify_admins, mock_notify
    ):
        # given
        character = EveCharacterFactory()
        user = UserMainBasicFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 0)

    def test_raise_error_when_no_valid_token_found_and_disable_character(
        self, mock_notify_admins, mock_notify
    ):
        # given
        eve_character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=eve_character)
        owner = OwnerFactory(user=user, characters=[eve_character])
        user.token_set.first().scopes.clear()  # token no longer valid
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        character = owner.characters.first()
        self.assertFalse(character.is_enabled)
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)

    def test_raise_error_when_character_no_longer_a_corporation_member_and_delete_it(
        self, mock_notify_admins, mock_notify
    ):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        new_corporation = EveCorporationInfoFactory()
        character.corporation_id = new_corporation.corporation_id
        character.corporation_name = new_corporation.corporation_name
        character.save()
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 0)

    def test_should_rotate_through_enabled_characters_for_notification(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = OwnerFactory(characters=False)
        character_1 = OwnerCharacterFactory(
            owner=owner,
            notifications_last_used_at=dt.datetime(2021, 1, 1, 1, 0, tzinfo=utc),
        )
        character_2 = OwnerCharacterFactory(
            owner=owner,
            notifications_last_used_at=dt.datetime(2021, 1, 1, 2, 0, tzinfo=utc),
        )
        character_3 = OwnerCharacterFactory(
            owner=owner,
            notifications_last_used_at=dt.datetime(2021, 1, 1, 3, 0, tzinfo=utc),
        )
        OwnerCharacterFactory(
            owner=owner, is_enabled=False
        )  # this one should be ignore
        tokens_received = []

        # when
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )

        # then
        self.assertListEqual(
            tokens_received,
            [
                character_1.character_id(),
                character_2.character_id(),
                character_3.character_id(),
                character_1.character_id(),
            ],
        )

    def test_should_rotate_through_characters_for_structures(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = OwnerFactory(characters=False)
        character_1 = OwnerCharacterFactory(
            owner=owner,
            structures_last_used_at=dt.datetime(2021, 1, 1, 3, 0, tzinfo=utc),
        )
        character_2 = OwnerCharacterFactory(
            owner=owner,
            structures_last_used_at=dt.datetime(2021, 1, 1, 1, 0, tzinfo=utc),
        )
        character_3 = OwnerCharacterFactory(
            owner=owner,
            structures_last_used_at=dt.datetime(2021, 1, 1, 2, 0, tzinfo=utc),
        )
        tokens_received = []

        # when
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )

        # then
        self.assertListEqual(
            tokens_received,
            [
                character_2.character_id(),
                character_3.character_id(),
                character_1.character_id(),
                character_2.character_id(),
            ],
        )

    def test_should_delete_invalid_characters_and_return_token_from_valid_char(
        self, mock_notify_admins, mock_notify
    ):
        # given
        character_1 = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character_1)
        owner = OwnerFactory(
            user=user,
            characters=[character_1],
            characters__notifications_last_used_at=dt.datetime(
                2021, 1, 1, 1, 2, tzinfo=utc
            ),
        )
        character_2 = EveCharacterFactory()  # invalid, because of different corporation
        OwnerCharacterFactory(
            owner=owner,
            eve_character=character_2,
            notifications_last_used_at=dt.datetime(2021, 1, 1, 1, 1, tzinfo=utc),
        )

        # when
        token = owner.fetch_token()

        # then
        self.assertIsInstance(token, Token)
        self.assertEqual(token.user, user)
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 1)


class TestOwnerCharacters(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory()

    def test_should_add_new_character(self):
        # given
        character = EveCharacterFactory(corporation=self.owner.corporation)
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        character_ownership = user.profile.main_character.character_ownership
        # when
        result = self.owner.add_character(character_ownership)
        # then
        self.assertIsInstance(result, OwnerCharacter)
        self.assertEqual(result.owner, self.owner)
        self.assertEqual(result.character_ownership, character_ownership)
        self.assertIsNone(result.notifications_last_used_at)
        character_ownership_pks = self.owner.characters.values_list(
            "character_ownership", flat=True
        )
        self.assertIn(character_ownership.pk, character_ownership_pks)
        self.assertEqual(character_ownership_pks.count(), 2)

    def test_should_not_overwrite_existing_characters(self):
        # given
        character = self.owner.characters.first()
        my_dt = datetime(year=2021, month=2, day=11, hour=12, tzinfo=utc)
        character.notifications_last_used_at = my_dt
        character.save()
        # when
        result = self.owner.add_character(character.character_ownership)
        # then
        self.assertIsInstance(result, OwnerCharacter)
        self.assertEqual(result.owner, self.owner)
        self.assertEqual(result.character_ownership, character.character_ownership)
        self.assertEqual(result.notifications_last_used_at, my_dt)

    def test_should_prevent_adding_character_from_other_corporation(self):
        # given
        user = UserMainDefaultOwnerFactory()
        character_ownership = user.profile.main_character.character_ownership
        # when
        with self.assertRaises(ValueError):
            self.owner.add_character(character_ownership)

    def test_should_count_enabled_characters_only(self):
        # given
        OwnerCharacterFactory(owner=self.owner, is_enabled=False)
        # when
        result = self.owner.valid_characters_count()
        # then
        self.assertEqual(result, 1)

    def test_should_count_characters_when_empty(self):
        # given
        owner = OwnerFactory(characters=False)
        # when
        result = owner.valid_characters_count()
        # then
        self.assertEqual(result, 0)

    def test_should_reset_character_when_re_adding(self):
        # given
        character: OwnerCharacter = self.owner.characters.first()
        character.is_enabled = False
        character.disabled_reason = "some reason"
        character.save()
        # when
        self.owner.add_character(character.character_ownership)
        # then
        character.refresh_from_db()
        self.assertTrue(character.is_enabled)
        self.assertFalse(character.disabled_reason)


@patch(MODULE_PATH + ".notify", spec=True)
@patch(MODULE_PATH + ".notify_admins", spec=True)
class TestOwnerDeleteCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory(characters=False)

    def test_should_delete_character_and_notify(self, mock_notify_admins, mock_notify):
        # given
        character = OwnerCharacterFactory(owner=self.owner)
        user = character.character_ownership.user

        # when
        self.owner.delete_character(character=character, reason="dummy error")

        # then
        self.assertEqual(self.owner.characters.count(), 0)
        self.assertTrue(mock_notify_admins.called)
        _, kwargs = mock_notify_admins.call_args
        self.assertIn("dummy error", kwargs["message"])
        self.assertEqual(kwargs["level"], "danger")
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertIn("dummy error", kwargs["message"])
        self.assertEqual(kwargs["user"], user)
        self.assertEqual(kwargs["level"], "warning")


@patch(MODULE_PATH + ".notify", spec=True)
@patch(MODULE_PATH + ".notify_admins", spec=True)
class TestOwnerDisableCharacters(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory(characters=False)

    def test_should_disable_character_and_notify(self, mock_notify_admins, mock_notify):
        # given
        character = OwnerCharacterFactory(owner=self.owner)
        user = character.character_ownership.user

        # when
        self.owner.disable_character(character=character, reason="dummy error")

        # then
        character.refresh_from_db()
        self.assertFalse(character.is_enabled)
        self.assertTrue(character.disabled_reason)
        self.assertTrue(mock_notify_admins.called)
        _, kwargs = mock_notify_admins.call_args
        self.assertIn("dummy error", kwargs["message"])
        self.assertEqual(kwargs["level"], "danger")
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertIn("dummy error", kwargs["message"])
        self.assertEqual(kwargs["user"], user)
        self.assertEqual(kwargs["level"], "warning")

    def test_should_not_disable_when_error_counter_above_zero(
        self, mock_notify_admins, mock_notify
    ):
        # given
        character = OwnerCharacterFactory(owner=self.owner)

        # when
        self.owner.disable_character(
            character=character, reason="dummy error", max_allowed_errors=1
        )
        # then
        character.refresh_from_db()
        self.assertTrue(character.is_enabled)
        self.assertEqual(character.error_count, 1)
        self.assertFalse(mock_notify_admins.called)
        self.assertFalse(mock_notify.called)



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structures.core.notification_types import NotificationType
from structures.models import FuelAlert, JumpFuelAlert, Notification, Structure, Webhook
from structures.tests.testdata.factories import (
    EveEntityCorporationFactory,
    FuelAlertConfigFactory,
    FuelAlertFactory,
    JumpFuelAlertConfigFactory,
    JumpGateFactory,
    NotificationFactory,
    OwnerFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    WebhookFactory,
)
from structures.tests.testdata.helpers import (
    load_eve_entities,
    load_notification_by_type,
    load_notification_entities,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.notifications"


@patch(MODULE_PATH + ".Webhook.send_message", spec=True)
class TestStructureFuelAlerts(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):  # Can not be setUpTestData due to conflict with redis client
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        cls.owner = OwnerFactory()
        load_notification_entities(cls.owner)
        cls.webhook = cls.owner.webhooks.first()

    def test_should_output_str(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(owner=self.owner, fuel_expires_at=None)
        alert = FuelAlertFactory(structure=structure, config=config, hours=36)
        # when
        result = str(alert)
        # then
        self.assertIsInstance(result, str)

    def test_should_send_fuel_notification_for_structure(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        obj = FuelAlert.objects.first()
        self.assertEqual(obj.hours, 36)

    def test_should_not_send_fuel_notification_for_structure_not_burning_fuel(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        StructureFactory(
            owner=self.owner, fuel_expires_at=now() - dt.timedelta(hours=2)
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)

    def test_should_not_send_fuel_notification_that_already_exists(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_message.reset_mock()
        FuelAlertFactory(structure=structure, config=config, hours=36)
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)
        self.assertEqual(FuelAlert.objects.count(), 1)

    def test_should_send_fuel_notification_for_starbase(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        StarbaseFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        obj = FuelAlert.objects.first()
        self.assertEqual(obj.hours, 36)

    def test_should_not_send_fuel_notification_for_starbase_not_burning_fuel(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        StarbaseFactory(
            owner=self.owner,
            state=Structure.State.POS_OFFLINE,
            fuel_expires_at=now() - dt.timedelta(hours=2),
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)

    def test_should_use_configured_ping_type_for_notifications(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(
            start=48,
            end=0,
            repeat=12,
            channel_ping_type=Webhook.PingType.EVERYONE,
        )
        StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        _, kwargs = mock_send_message.call_args
        self.assertIn("@everyone", kwargs["content"])

    def test_should_use_configured_level_for_notifications(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(
            start=48,
            end=0,
            repeat=12,
            color=Webhook.Color.SUCCESS,
        )
        StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        _, kwargs = mock_send_message.call_args
        embed = kwargs["embeds"][0]
        self.assertEqual(embed.color, Webhook.Color.SUCCESS)

    def test_should_send_fuel_notification_at_start(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=12, end=0, repeat=12)
        StructureFactory(
            owner=self.owner,
            fuel_expires_at=now() + dt.timedelta(hours=11, minutes=59, seconds=59),
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        obj = FuelAlert.objects.first()
        self.assertEqual(obj.hours, 12)

    def test_should_not_send_fuel_notifications_before_start(self, mock_send_message):
        # given
        config = FuelAlertConfigFactory(start=12, end=6, repeat=1)
        StructureFactory(
            owner=self.owner,
            fuel_expires_at=now() + dt.timedelta(hours=12, minutes=0, seconds=1),
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)

    def test_should_not_send_fuel_notifications_after_end(self, mock_send_message):
        # given
        config = FuelAlertConfigFactory(start=12, end=6, repeat=1)
        StructureFactory(
            owner=self.owner,
            fuel_expires_at=now() + dt.timedelta(hours=5, minutes=59, seconds=59),
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)

    def test_should_send_fuel_notification_at_start_when_repeat_is_0(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        config = FuelAlertConfigFactory(start=12, end=0, repeat=0)
        StructureFactory(
            owner=self.owner,
            fuel_expires_at=now() + dt.timedelta(hours=11, minutes=59, seconds=59),
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        obj = FuelAlert.objects.first()
        self.assertEqual(obj.hours, 12)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_send_structure_fuel_notification_to_configured_webhook_only(
        self, mock_send_to_webhook, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        webhook_2 = WebhookFactory(
            notification_types=[
                NotificationType.STRUCTURE_DESTROYED,
                NotificationType.TOWER_RESOURCE_ALERT_MSG,
            ]
        )
        self.owner.webhooks.add(webhook_2)
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        StructureFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_to_webhook.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertEqual(config.structure_fuel_alerts.count(), 1)
        self.assertEqual(mock_send_to_webhook.call_count, 1)
        args, _ = mock_send_to_webhook.call_args
        self.assertEqual(args[0], self.webhook)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_send_starbase_fuel_notification_to_configured_webhook_only(
        self, mock_send_to_webhook, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        webhook_2 = WebhookFactory(
            notification_types=[
                NotificationType.STRUCTURE_DESTROYED,
                NotificationType.STRUCTURE_FUEL_ALERT,
            ]
        )
        self.owner.webhooks.add(webhook_2)
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        StarbaseFactory(
            owner=self.owner, fuel_expires_at=now() + dt.timedelta(hours=25)
        )
        mock_send_to_webhook.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertEqual(config.structure_fuel_alerts.count(), 1)
        self.assertEqual(mock_send_to_webhook.call_count, 1)
        args, _ = mock_send_to_webhook.call_args
        self.assertEqual(args[0], self.webhook)

    def test_should_remove_alerts_when_config_changes_1(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(owner=self.owner)
        FuelAlertFactory(structure=structure, config=config, hours=36)
        # when
        config.start = 36
        config.save()
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    def test_should_remove_alerts_when_config_changes_2(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(owner=self.owner)
        FuelAlertFactory(structure=structure, config=config, hours=36)
        # when
        config.end = 2
        config.save()
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    def test_should_remove_alerts_when_config_changes_3(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(owner=self.owner)
        FuelAlertFactory(structure=structure, config=config, hours=36)
        # when
        config.repeat = 4
        config.save()
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    def test_should_keep_alerts_when_config_updated_without_change(
        self, mock_send_message
    ):
        # given
        mock_send_message.side_effect = RuntimeError
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure = StructureFactory(owner=self.owner)
        FuelAlertFactory(structure=structure, config=config, hours=36)
        # when
        config.save()
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 1)

    def test_should_return_correct_webhooks(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook_wrong_type = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_DESTROYED]
        )
        self.owner.webhooks.add(webhook_wrong_type)
        structure = StructureFactory(owner=self.owner)
        webhook_structure = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_FUEL_ALERT]
        )
        structure.webhooks.add(webhook_structure)
        webhook_inactive = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_FUEL_ALERT], is_active=False
        )
        self.owner.webhooks.add(webhook_inactive)
        WebhookFactory(notification_types=[NotificationType.STRUCTURE_FUEL_ALERT])
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        # when
        qs = config.relevant_webhooks()
        # then
        relevant_webhook_pks = qs.values_list("pk", flat=True)
        self.assertSetEqual(
            set(relevant_webhook_pks), {self.webhook.pk, webhook_structure.pk}
        )


@patch(MODULE_PATH + ".Webhook.send_message", spec=True)
class TestJumpFuelAlerts(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        EveEntityCorporationFactory(id=1000137, name="DED")

    def test_should_output_str(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        structure = JumpGateFactory()
        config = JumpFuelAlertConfigFactory(threshold=100)
        alert = structure.jump_fuel_alerts.create(config=config)
        # when
        result = str(alert)
        # then
        self.assertIsInstance(result, str)

    def test_should_send_fuel_notification_for_structure(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100)
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        alert = JumpFuelAlert.objects.first()
        self.assertEqual(alert.structure, structure)
        self.assertEqual(alert.config, config)

    def test_should_not_send_fuel_notification_for_structure_when_not_burning_fuel(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        JumpGateFactory(owner=owner, fuel_expires_at=None, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100)
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)

    def test_should_handle_no_fuel_situation(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        JumpGateFactory(owner=owner, jump_fuel_quantity=0)
        config = JumpFuelAlertConfigFactory(threshold=100)
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)

    def test_should_not_send_fuel_notification_that_already_exists(
        self, mock_send_message
    ):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100)
        alert = structure.jump_fuel_alerts.create(config=config)
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)
        self.assertEqual(structure.jump_fuel_alerts.count(), 1)
        self.assertEqual(structure.jump_fuel_alerts.first(), alert)

    def test_should_not_send_fuel_notification_if_above_threshold(
        self, mock_send_message
    ):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = JumpGateFactory(owner=owner, jump_fuel_quantity=101)
        config = JumpFuelAlertConfigFactory(threshold=100)
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertFalse(mock_send_message.called)
        self.assertEqual(structure.jump_fuel_alerts.count(), 0)

    def test_should_use_configured_ping_type_for_notifications(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(
            threshold=100, channel_ping_type=Webhook.PingType.EVERYONE
        )
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        _, kwargs = mock_send_message.call_args
        self.assertIn("@everyone", kwargs["content"])

    def test_should_use_configured_level_for_notifications(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100, color=Webhook.Color.SUCCESS)
        mock_send_message.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertTrue(mock_send_message.called)
        _, kwargs = mock_send_message.call_args
        embed = kwargs["embeds"][0]
        self.assertEqual(embed.color, Webhook.Color.SUCCESS)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_send_fuel_notification_to_configured_webhook_only(
        self, mock_send_to_webhook, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        webhook_other = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_UNDER_ATTACK]
        )
        owner = OwnerFactory(webhooks=[webhook, webhook_other])
        JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100)
        mock_send_to_webhook.reset_mock()
        # when
        config.send_new_notifications()
        # then
        self.assertEqual(config.jump_fuel_alerts.count(), 1)
        self.assertEqual(mock_send_to_webhook.call_count, 1)
        args, _ = mock_send_to_webhook.call_args
        self.assertEqual(args[0], webhook)

    def test_should_remove_alerts_when_config_changes(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100)
        structure.jump_fuel_alerts.create(config=config)
        # when
        config.threshold = 50
        config.save()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 0)

    def test_should_keep_alerts_when_config_updated_without_change(
        self, mock_send_message
    ):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = JumpGateFactory(owner=owner, jump_fuel_quantity=99)
        config = JumpFuelAlertConfigFactory(threshold=100)
        structure.jump_fuel_alerts.create(config=config)
        # when
        config.save()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 1)

    def test_should_return_correct_webhooks(self, mock_send_message):
        # given
        mock_send_message.side_effect = RuntimeError
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        webhook_wrong_type = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_DESTROYED]
        )
        webhook_inactive = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT],
            is_active=False,
        )
        owner = OwnerFactory(webhooks=[webhook, webhook_wrong_type, webhook_inactive])
        structure = JumpGateFactory(owner=owner)
        webhook_structure = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT]
        )
        structure.webhooks.add(webhook_structure)
        WebhookFactory(notification_types=[NotificationType.STRUCTURE_JUMP_FUEL_ALERT])
        config = JumpFuelAlertConfigFactory(threshold=100)
        # when
        qs = config.relevant_webhooks()
        # then
        relevant_webhook_pks = qs.values_list("pk", flat=True)
        self.assertSetEqual(
            set(relevant_webhook_pks), {webhook.pk, webhook_structure.pk}
        )


class TestNotificationRelatedStructures(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        cls.owner = OwnerFactory()

    def test_related_structures_for_structure_notifications(self):
        # given
        structure = StructureFactory(owner=self.owner, id=1000000000001)
        for notif_type in [
            NotificationType.STRUCTURE_ONLINE,
            NotificationType.STRUCTURE_FUEL_ALERT,
            NotificationType.STRUCTURE_SERVICES_OFFLINE,
            NotificationType.STRUCTURE_WENT_LOW_POWER,
            NotificationType.STRUCTURE_WENT_HIGH_POWER,
            NotificationType.STRUCTURE_UNANCHORING,
            NotificationType.STRUCTURE_UNDER_ATTACK,
            NotificationType.STRUCTURE_LOST_SHIELD,
            NotificationType.STRUCTURE_LOST_ARMOR,
            NotificationType.STRUCTURE_DESTROYED,
            NotificationType.OWNERSHIP_TRANSFERRED,
            NotificationType.STRUCTURE_ANCHORING,
        ]:
            with self.subTest(notif_type=notif_type):
                notif = load_notification_by_type(
                    owner=self.owner, notif_type=notif_type
                )
                # when
                result_qs = notif.calc_related_structures()
                # then
                self.assertQuerysetEqual(
                    result_qs, Structure.objects.filter(id=structure.id)
                )

    def test_related_structures_for_moon_notifications(self):
        # given
        structure = StructureFactory(owner=self.owner, id=1000000000002)
        for notif_type in [
            NotificationType.MOONMINING_EXTRACTION_STARTED,
            NotificationType.MOONMINING_EXTRACTION_FINISHED,
            NotificationType.MOONMINING_AUTOMATIC_FRACTURE,
            NotificationType.MOONMINING_EXTRACTION_CANCELLED,
            NotificationType.MOONMINING_LASER_FIRED,
        ]:
            with self.subTest(notif_type=notif_type):
                notif = load_notification_by_type(
                    owner=self.owner, notif_type=notif_type
                )
                # when
                result_qs = notif.calc_related_structures()
                # then
                self.assertQuerysetEqual(
                    result_qs, Structure.objects.filter(id=structure.id)
                )

    def test_related_structures_for_orbital_notifications(self):
        # given
        structure = PocoFactory(owner=self.owner, eve_planet_id=40161469)
        for notif_type in [
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.ORBITAL_REINFORCED,
        ]:
            with self.subTest(notif_type=notif_type):
                notif = load_notification_by_type(
                    owner=self.owner, notif_type=notif_type
                )
                # when
                result_qs = notif.calc_related_structures()
                # then
                self.assertQuerysetEqual(
                    result_qs, Structure.objects.filter(id=structure.id)
                )

    def test_related_structures_for_tower_notifications(self):
        # given
        structure = StarbaseFactory(owner=self.owner, eve_moon_id=40161465)
        for notif_type in [
            NotificationType.TOWER_ALERT_MSG,
            NotificationType.TOWER_RESOURCE_ALERT_MSG,
        ]:
            with self.subTest(notif_type=notif_type):
                notif = load_notification_by_type(
                    owner=self.owner, notif_type=notif_type
                )
                # when
                result_qs = notif.calc_related_structures()
                # then
                self.assertQuerysetEqual(
                    result_qs, Structure.objects.filter(id=structure.id)
                )

    def test_related_structures_for_generated_notifications(self):
        # given
        structure = StarbaseFactory(owner=self.owner, eve_moon_id=40161465)
        for notif_type in [
            NotificationType.STRUCTURE_JUMP_FUEL_ALERT,
            NotificationType.STRUCTURE_REFUELED_EXTRA,
            NotificationType.TOWER_REFUELED_EXTRA,
        ]:
            with self.subTest(notif_type=notif_type):
                notif = Notification.create_from_structure(
                    structure, notif_type=notif_type
                )
                # when
                result_qs = notif.calc_related_structures()
                # then
                self.assertQuerysetEqual(
                    result_qs, Structure.objects.filter(id=structure.id)
                )

    def test_should_update_related_structure_when_it_exists(self):
        # given
        structure = StructureFactory(owner=self.owner)
        notif = NotificationFactory(owner=self.owner)
        # when
        with patch(MODULE_PATH + ".Notification.calc_related_structures") as m:
            m.return_value = Structure.objects.filter(id=structure.id)
            result = notif.update_related_structures()
        # then
        structure_ids = notif.structures.values_list("id", flat=True)
        self.assertSetEqual(set(structure_ids), {structure.id})
        self.assertTrue(result)

    def test_should_not_update_related_structure_when_not_found(self):
        # given
        notif = NotificationFactory(owner=self.owner)
        # when
        with patch(MODULE_PATH + ".Notification.calc_related_structures") as m:
            m.return_value = Structure.objects.none()
            result = notif.update_related_structures()
        # then
        self.assertFalse(result)

    def test_should_return_empty_qs_when_structure_id_is_missing(self):
        # given
        notif = NotificationFactory(
            notif_type=NotificationType.OWNERSHIP_TRANSFERRED.value,
            text_from_dict={
                "charID": 1001,
                "newOwnerCorpID": 2002,
                "oldOwnerCorpID": 2001,
                "solarSystemID": 30002537,
                "structureName": "Test Structure Alpha",
                "structureTypeID": 35832,
            },
        )
        # when
        result_qs = notif.calc_related_structures()
        # then
        self.assertEqual(list(result_qs), [])

    def test_should_return_empty_qs_when_all_structure_info_is_missing(self):
        # given
        notif = NotificationFactory(
            notif_type=NotificationType.STRUCTURE_REINFORCEMENT_CHANGED.value,
            text_from_dict={
                "hour": 19,
                "numStructures": 1,
                "timestamp": 132141703753688216,
                "weekday": 255,
            },
        )
        # when
        result_qs = notif.calc_related_structures()
        # then
        self.assertEqual(list(result_qs), [])

    def test_should_return_empty_qs_when_planet_id_is_missing(self):
        # given
        notif = NotificationFactory(
            notif_type=NotificationType.ORBITAL_ATTACKED.value,
            text_from_dict={
                "aggressorCorpID": 2011,
                "aggressorID": 1011,
                "planetTypeID": 2016,
                "shieldLevel": 0.9821850015653375,
                "solarSystemID": 30002537,
                "typeID": 2233,
            },
        )
        # when
        result_qs = notif.calc_related_structures()
        # then
        self.assertEqual(list(result_qs), [])

    def test_should_return_empty_qs_when_type_id_is_missing(self):
        # given
        notif = NotificationFactory(
            notif_type=NotificationType.ORBITAL_ATTACKED.value,
            text_from_dict={
                "aggressorCorpID": 2011,
                "aggressorID": 1011,
                "planetID": 40161469,
                "planetTypeID": 2016,
                "shieldLevel": 0.9821850015653375,
                "solarSystemID": 30002537,
            },
        )
        # when
        result_qs = notif.calc_related_structures()
        # then
        self.assertEqual(list(result_qs), [])

    def test_should_return_empty_qs_when_moon_id_is_missing(self):
        # given
        notif = NotificationFactory(
            notif_type=NotificationType.TOWER_ALERT_MSG.value,
            text_from_dict={
                "aggressorAllianceID": 3011,
                "aggressorCorpID": 2011,
                "aggressorID": 1011,
                "armorValue": 0.6950949076033535,
                "hullValue": 1.0,
                "shieldValue": 0.3950949076033535,
                "solarSystemID": 30002537,
                "typeID": 16213,
            },
        )
        # when
        result_qs = notif.calc_related_structures()
        # then
        self.assertEqual(list(result_qs), [])

    def test_should_return_empty_qs_when_moon_type_id_is_missing(self):
        # given
        notif = NotificationFactory(
            notif_type=NotificationType.TOWER_ALERT_MSG.value,
            text_from_dict={
                "aggressorAllianceID": 3011,
                "aggressorCorpID": 2011,
                "aggressorID": 1011,
                "armorValue": 0.6950949076033535,
                "hullValue": 1.0,
                "moonID": 40161465,
                "shieldValue": 0.3950949076033535,
                "solarSystemID": 30002537,
            },
        )
        # when
        result_qs = notif.calc_related_structures()
        # then
        self.assertEqual(list(result_qs), [])



"""Billing embeds."""

# pylint: disable=missing-class-docstring

import re
from typing import Optional

import dhooks_lite

from django.conf import settings
from django.utils.translation import gettext as _

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.urls import reverse_absolute, static_file_absolute_url

from structures import __title__
from structures.core.notification_types import NotificationType
from structures.helpers import get_or_create_eve_entity, is_absolute_url
from structures.models.notifications import Notification, NotificationBase, Webhook

from .helpers import gen_alliance_link, gen_corporation_link, target_datetime_formatted

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class NotificationBaseEmbed:
    """Base class for all notification embeds.

    You must subclass this class to create an embed for a notification type.
    At least title and description must be defined in the subclass.
    """

    ICON_DEFAULT_SIZE = 64

    def __init__(self, notification: Notification) -> None:
        if not isinstance(notification, NotificationBase):
            raise TypeError("notification must be of type Notification")
        self._notification = notification
        self._data = notification.parsed_text()
        self._title = ""
        self._description = ""
        self._color = None
        self._thumbnail = None
        self._ping_type = None

    def __str__(self) -> str:
        return str(self.notification)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(notification={self.notification!r})"

    @property
    def notification(self) -> Notification:
        """Return notification object this embed is created from."""
        return self._notification

    @property
    def ping_type(self) -> Optional[Webhook.PingType]:
        """Return Ping Type of the related notification."""
        return self._ping_type

    def compile_damage_text(self, field_postfix: str, factor: int = 1) -> str:
        """Compile damage text for Structures and POSes"""
        damage_labels = [
            ("shield", _("shield")),
            ("armor", _("armor")),
            ("hull", _("hull")),
        ]
        damage_parts = []
        for prop in damage_labels:
            field_name = f"{prop[0]}{field_postfix}"
            if field_name in self._data:
                label = prop[1]
                value = self._data[field_name] * factor
                damage_parts.append(f"{label}: {value:.1f}%")
        damage_text = " | ".join(damage_parts)
        return damage_text

    def gen_aggressor_link(self) -> str:
        """Returns the aggressor link from a parsed_text for POS and POCOs only."""
        if key := self._data.get("aggressorAllianceID"):
            key = "aggressorAllianceID"
        elif self._data.get("aggressorCorpID"):
            key = "aggressorCorpID"
        elif self._data.get("aggressorID"):
            key = "aggressorID"
        else:
            return "(Unknown aggressor)"
        entity = get_or_create_eve_entity(id=self._data[key])
        return Webhook.create_link(entity.name, entity.profile_url)

    def gen_attacker_link(self) -> str:
        """Returns the attacker link from a parsed_text for Upwell structures only."""
        if name := self._data.get("allianceName"):
            return gen_alliance_link(name)

        if name := self._data.get("corpName"):
            return gen_corporation_link(name)

        return _("(unknown)")

    def fuel_expires_target_date(self) -> str:
        """Return calculated target date when fuel expires. Returns '?' when no data."""
        if self._structure and self._structure.fuel_expires_at:
            return target_datetime_formatted(self._structure.fuel_expires_at)
        return "?"

    def generate_embed(self) -> dhooks_lite.Embed:
        """Returns generated Discord embed for this object.

        Will use custom color for embeds if self.notification has the
        property "color_override" defined

        Will use custom ping type if self.notification has the
        property "ping_type_override" defined

        """
        corporation = self.notification.owner.corporation
        if self.notification.is_alliance_level and corporation.alliance:
            author_name = corporation.alliance.alliance_name
            author_url = corporation.alliance.logo_url(size=self.ICON_DEFAULT_SIZE)
        else:
            author_name = corporation.corporation_name
            author_url = corporation.logo_url(size=self.ICON_DEFAULT_SIZE)
        app_url = reverse_absolute("structures:index")
        app_url = app_url if is_absolute_url(app_url) else None
        author = dhooks_lite.Author(name=author_name, icon_url=author_url, url=app_url)
        if self.notification.color_override:
            self._color = self.notification.color_override
        if self.notification.ping_type_override:
            self._ping_type = self.notification.ping_type_override
        elif self._color == Webhook.Color.DANGER:
            self._ping_type = Webhook.PingType.EVERYONE
        elif self._color == Webhook.Color.WARNING:
            self._ping_type = Webhook.PingType.HERE
        else:
            self._ping_type = Webhook.PingType.NONE
        if self.notification.is_generated:
            footer_text = __title__
            footer_icon_url = static_file_absolute_url(
                "structures/img/structures_logo.png"
            )
        else:
            footer_text = "Eve Online"
            footer_icon_url = static_file_absolute_url(
                "structures/img/eve_symbol_128.png"
            )
        if settings.DEBUG:
            my_text = (
                self.notification.notification_id
                if not self.notification.is_generated
                else "GENERATED"
            )
            footer_text += f" #{my_text}"
        footer_icon_url = footer_icon_url if is_absolute_url(footer_icon_url) else None
        footer = dhooks_lite.Footer(text=footer_text, icon_url=footer_icon_url)
        max_description = dhooks_lite.Embed.MAX_DESCRIPTION
        if self._description and len(self._description) > max_description:
            logger.warning(
                "%s: Description of notification is too long: %s",
                self,
                self._description,
            )
            self._description = self._description[:max_description]
        max_title = dhooks_lite.Embed.MAX_TITLE
        if self._title and len(self._title) > max_title:
            logger.warning(
                "%s: Title of notification is too long: %s", self, self._title
            )
            self._title = self._title[:max_title]
        return dhooks_lite.Embed(
            author=author,
            color=self._color,
            description=self._description,
            footer=footer,
            timestamp=self.notification.timestamp,
            title=self._title,
            thumbnail=self._thumbnail,
        )

    # pylint: disable = too-many-locals
    @staticmethod
    def create(notif: "NotificationBase") -> "NotificationBaseEmbed":
        """Creates a new instance of the respective subclass for given Notification."""

        from .billing_embeds import (
            NotificationBillingBillOutOfMoneyMsg,
            NotificationBillingIHubBillAboutToExpire,
            NotificationBillingIHubDestroyedByBillFailure,
            NotificationCorpAllBillMsg,
        )
        from .corporate_embeds import (
            NotificationCharAppAcceptMsg,
            NotificationCharAppRejectMsg,
            NotificationCharAppWithdrawMsg,
            NotificationCharLeftCorpMsg,
            NotificationCorpAppInvitedMsg,
            NotificationCorpAppNewMsg,
            NotificationCorpAppRejectCustomMsg,
        )
        from .moonmining_embeds import (
            NotificationMoonminningAutomaticFracture,
            NotificationMoonminningExtractionCanceled,
            NotificationMoonminningExtractionFinished,
            NotificationMoonminningExtractionStarted,
            NotificationMoonminningLaserFired,
        )
        from .orbital_embeds import (
            NotificationOrbitalAttacked,
            NotificationOrbitalReinforced,
        )
        from .skyhook_embeds import (
            NotificationSkyhookDeployed,
            NotificationSkyhookDestroyed,
            NotificationSkyhookLostShield,
            NotificationSkyhookOnline,
            NotificationSkyhookUnderAttack,
        )
        from .sov_embeds import (
            NotificationSovAllAnchoringMsg,
            NotificationSovAllClaimAcquiredMsg,
            NotificationSovAllClaimLostMsg,
            NotificationSovCommandNodeEventStarted,
            NotificationSovEntosisCaptureStarted,
            NotificationSovStructureDestroyed,
            NotificationSovStructureReinforced,
        )
        from .structures_embeds import (
            NotificationStructureAnchoring,
            NotificationStructureDestroyed,
            NotificationStructureFuelAlert,
            NotificationStructureJumpFuelAlert,
            NotificationStructureLostArmor,
            NotificationStructureLostShield,
            NotificationStructureLowReagentsAlert,
            NotificationStructureNoReagentsAlert,
            NotificationStructureOnline,
            NotificationStructureOwnershipTransferred,
            NotificationStructureRefueledExtra,
            NotificationStructureReinforceChange,
            NotificationStructureServicesOffline,
            NotificationStructureUnanchoring,
            NotificationStructureUnderAttack,
            NotificationStructureWentHighPower,
            NotificationStructureWentLowPower,
        )
        from .tower_embeds import (
            NotificationTowerAlertMsg,
            NotificationTowerRefueledExtra,
            NotificationTowerReinforcedExtra,
            NotificationTowerResourceAlertMsg,
        )
        from .war_embeds import (
            NotificationAcceptedAlly,
            NotificationAllWarCorpJoinedAllianceMsg,
            NotificationAllWarSurrenderMsg,
            NotificationAllyJoinedWarMsg,
            NotificationCorpWarSurrenderMsg,
            NotificationDeclareWar,
            NotificationMercOfferedNegotiationMsg,
            NotificationMercOfferRetractedMsg,
            NotificationOfferedSurrender,
            NotificationOfferedToAlly,
            NotificationWarAdopted,
            NotificationWarCorporationBecameEligible,
            NotificationWarCorporationNoLongerEligible,
            NotificationWarDeclared,
            NotificationWarHQRemovedFromSpace,
            NotificationWarInherited,
            NotificationWarInvalid,
            NotificationWarRetractedByConcord,
            NotificationWarSurrenderOfferMsg,
        )

        if not isinstance(notif, NotificationBase):
            raise TypeError("notification must be of type NotificationBase")

        NT = NotificationType
        notif_type_2_class = {
            # Billing
            NT.BILLING_BILL_OUT_OF_MONEY_MSG: NotificationBillingBillOutOfMoneyMsg,
            NT.BILLING_CORP_ALL_BILL_MSG: NotificationCorpAllBillMsg,
            NT.BILLING_I_HUB_BILL_ABOUT_TO_EXPIRE: NotificationBillingIHubBillAboutToExpire,
            NT.BILLING_I_HUB_DESTROYED_BY_BILL_FAILURE: NotificationBillingIHubDestroyedByBillFailure,
            # character
            NT.CHAR_APP_ACCEPT_MSG: NotificationCharAppAcceptMsg,
            NT.CHAR_APP_WITHDRAW_MSG: NotificationCharAppWithdrawMsg,
            NT.CHAR_LEFT_CORP_MSG: NotificationCharLeftCorpMsg,
            NT.CORP_APP_INVITED_MSG: NotificationCorpAppInvitedMsg,
            NT.CORP_APP_NEW_MSG: NotificationCorpAppNewMsg,
            NT.CORP_APP_REJECT_CUSTOM_MSG: NotificationCorpAppRejectCustomMsg,
            NT.CORP_APP_REJECT_MSG: NotificationCharAppRejectMsg,
            # moonmining
            NT.MOONMINING_AUTOMATIC_FRACTURE: NotificationMoonminningAutomaticFracture,
            NT.MOONMINING_EXTRACTION_CANCELLED: NotificationMoonminningExtractionCanceled,
            NT.MOONMINING_EXTRACTION_FINISHED: NotificationMoonminningExtractionFinished,
            NT.MOONMINING_EXTRACTION_STARTED: NotificationMoonminningExtractionStarted,
            NT.MOONMINING_LASER_FIRED: NotificationMoonminningLaserFired,
            # Orbitals
            NT.ORBITAL_ATTACKED: NotificationOrbitalAttacked,
            NT.ORBITAL_REINFORCED: NotificationOrbitalReinforced,
            # Sov
            NT.SOV_ALL_ANCHORING_MSG: NotificationSovAllAnchoringMsg,
            NT.SOV_ALL_CLAIM_ACQUIRED_MSG: NotificationSovAllClaimAcquiredMsg,
            NT.SOV_ALL_CLAIM_LOST_MSG: NotificationSovAllClaimLostMsg,
            NT.SOV_COMMAND_NODE_EVENT_STARTED: NotificationSovCommandNodeEventStarted,
            NT.SOV_ENTOSIS_CAPTURE_STARTED: NotificationSovEntosisCaptureStarted,
            NT.SOV_STRUCTURE_DESTROYED: NotificationSovStructureDestroyed,
            NT.SOV_STRUCTURE_REINFORCED: NotificationSovStructureReinforced,
            # Towers
            NT.TOWER_ALERT_MSG: NotificationTowerAlertMsg,
            NT.TOWER_REFUELED_EXTRA: NotificationTowerRefueledExtra,
            NT.TOWER_REINFORCED_EXTRA: NotificationTowerReinforcedExtra,
            NT.TOWER_RESOURCE_ALERT_MSG: NotificationTowerResourceAlertMsg,
            # Skyhooks
            NT.SKYHOOK_DEPLOYED: NotificationSkyhookDeployed,
            NT.SKYHOOK_DESTROYED: NotificationSkyhookDestroyed,
            NT.SKYHOOK_LOST_SHIELDS: NotificationSkyhookLostShield,
            NT.SKYHOOK_ONLINE: NotificationSkyhookOnline,
            NT.SKYHOOK_UNDER_ATTACK: NotificationSkyhookUnderAttack,
            # Upwell structures
            NT.OWNERSHIP_TRANSFERRED: NotificationStructureOwnershipTransferred,
            NT.STRUCTURE_ANCHORING: NotificationStructureAnchoring,
            NT.STRUCTURE_DESTROYED: NotificationStructureDestroyed,
            NT.STRUCTURE_FUEL_ALERT: NotificationStructureFuelAlert,
            NT.STRUCTURE_JUMP_FUEL_ALERT: NotificationStructureJumpFuelAlert,
            NT.STRUCTURE_LOST_ARMOR: NotificationStructureLostArmor,
            NT.STRUCTURE_LOST_SHIELD: NotificationStructureLostShield,
            NT.STRUCTURE_LOW_REAGENTS_ALERT: NotificationStructureLowReagentsAlert,
            NT.STRUCTURE_NO_REAGENTS_ALERT: NotificationStructureNoReagentsAlert,
            NT.STRUCTURE_ONLINE: NotificationStructureOnline,
            NT.STRUCTURE_REFUELED_EXTRA: NotificationStructureRefueledExtra,
            NT.STRUCTURE_REINFORCEMENT_CHANGED: NotificationStructureReinforceChange,
            NT.STRUCTURE_SERVICES_OFFLINE: NotificationStructureServicesOffline,
            NT.STRUCTURE_UNANCHORING: NotificationStructureUnanchoring,
            NT.STRUCTURE_UNDER_ATTACK: NotificationStructureUnderAttack,
            NT.STRUCTURE_WENT_HIGH_POWER: NotificationStructureWentHighPower,
            NT.STRUCTURE_WENT_LOW_POWER: NotificationStructureWentLowPower,
            # War
            NT.WAR_ACCEPTED_ALLY: NotificationAcceptedAlly,
            NT.WAR_ALL_WAR_CORP_JOINED_ALLIANCE_MSG: NotificationAllWarCorpJoinedAllianceMsg,
            NT.WAR_ALL_WAR_SURRENDER_MSG: NotificationAllWarSurrenderMsg,
            NT.WAR_ALLY_JOINED_WAR_AGGRESSOR_MSG: NotificationAllyJoinedWarMsg,
            NT.WAR_ALLY_JOINED_WAR_ALLY_MSG: NotificationAllyJoinedWarMsg,
            NT.WAR_ALLY_JOINED_WAR_DEFENDER_MSG: NotificationAllyJoinedWarMsg,
            NT.WAR_CORP_WAR_SURRENDER_MSG: NotificationCorpWarSurrenderMsg,
            NT.WAR_CORPORATION_BECAME_ELIGIBLE: NotificationWarCorporationBecameEligible,
            NT.WAR_CORPORATION_NO_LONGER_ELIGIBLE: NotificationWarCorporationNoLongerEligible,
            NT.WAR_DECLARE_WAR: NotificationDeclareWar,
            NT.WAR_HQ_REMOVED_FROM_SPACE: NotificationWarHQRemovedFromSpace,
            NT.WAR_INVALID: NotificationWarInvalid,
            NT.WAR_MERC_OFFER_RETRACTED_MSG: NotificationMercOfferRetractedMsg,
            NT.WAR_MERC_OFFERED_NEGOTIATION_MSG: NotificationMercOfferedNegotiationMsg,
            NT.WAR_OFFERED_SURRENDER: NotificationOfferedSurrender,
            NT.WAR_OFFERED_TO_ALLY: NotificationOfferedToAlly,
            NT.WAR_WAR_ADOPTED: NotificationWarAdopted,
            NT.WAR_WAR_DECLARED: NotificationWarDeclared,
            NT.WAR_WAR_INHERITED: NotificationWarInherited,
            NT.WAR_WAR_RETRACTED_BY_CONCORD: NotificationWarRetractedByConcord,
            NT.WAR_WAR_SURRENDER_OFFER_MSG: NotificationWarSurrenderOfferMsg,
        }
        try:
            notif_class = notif_type_2_class[notif.notif_type]
        except KeyError:
            return NotificationGenericEmbed(notif)

        return notif_class(notif)


class NotificationGenericEmbed(NotificationBaseEmbed):
    """A generic embed for undefined notifs."""

    def __init__(self, notif: Notification) -> None:
        super().__init__(notif)
        self._title = re.sub(
            r"((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))", r" \1", notif.notif_type
        )
        self._color = Webhook.Color.INFO
        self._thumbnail = dhooks_lite.Thumbnail(
            notif.sender.icon_url(size=self.ICON_DEFAULT_SIZE)
        )



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structures.core.notification_types import NotificationType
from structures.models import Notification, Webhook
from structures.tests.testdata.factories import (
    EveCorporationInfoFactory,
    EveEntityCorporationFactory,
    NotificationFactory,
    OwnerFactory,
    StarbaseFactory,
    StructureFactory,
    WebhookFactory,
)
from structures.tests.testdata.helpers import (
    clone_notification,
    load_eve_entities,
    load_notification_entities,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.notifications"


class TestNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        cls.owner = OwnerFactory()
        load_notification_entities(cls.owner)

    def test_str(self):
        # given
        obj = Notification.objects.get(notification_id=1000000403)
        # when/then
        self.assertEqual(str(obj), "1000000403:MoonminingExtractionFinished")

    def test_get_parsed_text(self):
        # given
        obj = Notification.objects.get(notification_id=1000000404)
        # when
        parsed_text = obj.parsed_text()
        # then
        self.assertEqual(parsed_text["autoTime"], 132186924601059151)
        self.assertEqual(parsed_text["structureName"], "Dummy")
        self.assertEqual(parsed_text["solarSystemID"], 30002537)

    def test_is_npc_attacking(self):
        for notification_id, expected in [
            (1000000509, False),
            (1000010509, True),
            (1000010601, True),
        ]:
            obj = Notification.objects.get(notification_id=notification_id)
            with self.subTest(notification=str(obj)):
                self.assertIs(obj.is_npc_attacking(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_REPORT_NPC_ATTACKS", True)
    def test_filter_npc_attacks_1(self):
        # NPC reporting allowed and not a NPC attacker
        obj = Notification.objects.get(notification_id=1000000509)
        self.assertFalse(obj.filter_for_npc_attacks())

        # NPC reporting allowed and a NPC attacker
        obj = Notification.objects.get(notification_id=1000010509)
        self.assertFalse(obj.filter_for_npc_attacks())

    @patch(MODULE_PATH + ".STRUCTURES_REPORT_NPC_ATTACKS", False)
    def test_filter_npc_attacks_2(self):
        # NPC reporting not allowed and not a NPC attacker
        obj = Notification.objects.get(notification_id=1000000509)
        self.assertFalse(obj.filter_for_npc_attacks())

        # NPC reporting not allowed and a NPC attacker
        obj = Notification.objects.get(notification_id=1000010509)
        self.assertTrue(obj.filter_for_npc_attacks())

    def test_can_be_rendered_1(self):
        for notif_type in NotificationType.values:
            with self.subTest(notification_type=notif_type):
                notif = Notification.objects.filter(notif_type=notif_type).first()
                if notif:
                    self.assertTrue(notif.can_be_rendered)

    def test_can_be_rendered_2(self):
        structure = StructureFactory(owner=self.owner, id=1000000000001)
        for notif_type in [
            NotificationType.STRUCTURE_REFUELED_EXTRA,
            NotificationType.TOWER_REFUELED_EXTRA,
        ]:
            with self.subTest(notification_type=notif_type):
                notif = Notification.create_from_structure(structure, notif_type)
                if notif:
                    self.assertTrue(notif.can_be_rendered)

    def test_can_be_rendered_3(self):
        for notif_type in ["UnknownNotificationType"]:
            with self.subTest(notification_type=notif_type):
                notif = Notification.objects.filter(notif_type=notif_type).first()
                if notif:
                    self.assertFalse(notif.can_be_rendered)


class TestNotificationFilterForAllianceLevel(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):  # TODO: Refactor so it works with setUpTestData()
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        cls.owner = OwnerFactory()
        load_notification_entities(cls.owner)

    def test_should_not_filter_non_alliance_notifications_1(self):
        # given
        self.owner.is_alliance_main = False
        self.owner.save()
        notifs = self.owner.notification_set.exclude(
            notif_type__in=NotificationType.relevant_for_alliance_level()
        )
        # when/then
        for notif in notifs:
            with self.subTest(notif=str(notif)):
                self.assertFalse(notif.filter_for_alliance_level())

    def test_should_not_filter_non_alliance_notifications_2(self):
        # given
        self.owner.is_alliance_main = True
        self.owner.save()
        notifs = self.owner.notification_set.exclude(
            notif_type__in=NotificationType.relevant_for_alliance_level()
        )
        # when/then
        for notif in notifs:
            with self.subTest(notif=str(notif)):
                self.assertFalse(notif.filter_for_alliance_level())

    def test_should_filter_alliance_notifications(self):
        # given
        self.owner.is_alliance_main = False
        self.owner.save()
        notifs = self.owner.notification_set.filter(
            notif_type__in=NotificationType.relevant_for_alliance_level()
        )
        # when/then
        for notif in notifs:
            with self.subTest(notif=str(notif)):
                self.assertTrue(notif.filter_for_alliance_level())

    def test_should_not_filter_alliance_notifications_1(self):
        # given
        self.owner.is_alliance_main = True
        self.owner.save()
        notifs = self.owner.notification_set.filter(
            notif_type__in=NotificationType.relevant_for_alliance_level()
        )
        # when/then
        for notif in notifs:
            with self.subTest(notif=str(notif)):
                self.assertFalse(notif.filter_for_alliance_level())

    def test_should_not_filter_alliance_notifications_2(self):
        # given
        self.owner.is_alliance_main = True
        self.owner.save()
        notifs = self.owner.notification_set.filter(
            notif_type__in=NotificationType.relevant_for_alliance_level()
        )
        # when/then
        for notif in notifs:
            with self.subTest(notif=str(notif)):
                self.assertFalse(notif.filter_for_alliance_level())

    def test_should_not_filter_alliance_notifications_3(self):
        # given
        corporation = EveCorporationInfoFactory(create_alliance=False)
        owner = OwnerFactory(corporation=corporation, is_alliance_main=True)
        load_notification_entities(owner)
        notifs = self.owner.notification_set.filter(
            notif_type__in=NotificationType.relevant_for_alliance_level()
        )
        # when/then
        for notif in notifs:
            with self.subTest(notif=str(notif)):
                self.assertFalse(notif.filter_for_alliance_level())


class TestNotificationCreateFromStructure(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        EveEntityCorporationFactory(id=1000137, name="DED")
        cls.owner = OwnerFactory()

    def test_should_create_notification_for_structure_fuel_alerts(self):
        # given
        structure = StructureFactory(owner=self.owner)
        # when
        notif = Notification.create_from_structure(
            structure, notif_type=NotificationType.STRUCTURE_FUEL_ALERT
        )
        # then
        self.assertIsInstance(notif, Notification)
        self.assertTrue(notif.is_temporary)
        self.assertAlmostEqual(notif.timestamp, now(), delta=dt.timedelta(seconds=10))
        self.assertAlmostEqual(
            notif.last_updated, now(), delta=dt.timedelta(seconds=10)
        )
        self.assertEqual(notif.owner, structure.owner)
        self.assertEqual(notif.sender_id, 1000137)
        self.assertEqual(notif.notif_type, NotificationType.STRUCTURE_FUEL_ALERT)

    def test_should_create_notification_for_tower_fuel_alerts(self):
        # given
        structure = StarbaseFactory(owner=self.owner)
        # when
        notif = Notification.create_from_structure(
            structure, notif_type=NotificationType.TOWER_RESOURCE_ALERT_MSG
        )
        # then
        self.assertIsInstance(notif, Notification)
        self.assertTrue(notif.is_temporary)
        self.assertAlmostEqual(notif.timestamp, now(), delta=dt.timedelta(seconds=10))
        self.assertAlmostEqual(
            notif.last_updated, now(), delta=dt.timedelta(seconds=10)
        )
        self.assertEqual(notif.owner, structure.owner)
        self.assertEqual(notif.sender_id, 1000137)
        self.assertEqual(notif.notif_type, NotificationType.TOWER_RESOURCE_ALERT_MSG)

    def test_should_create_notification_with_additional_params(self):
        # given
        structure = StructureFactory(owner=self.owner)
        # when
        notif = Notification.create_from_structure(
            structure, notif_type=NotificationType.STRUCTURE_FUEL_ALERT, is_read=True
        )
        # then
        self.assertIsInstance(notif, Notification)
        self.assertEqual(notif.notif_type, NotificationType.STRUCTURE_FUEL_ALERT)
        self.assertTrue(notif.is_read)

    def test_should_raise_error_when_text_is_missing(self):
        # given
        structure = StructureFactory(owner=self.owner)
        # when/then
        with self.assertRaises(ValueError):
            Notification.create_from_structure(
                structure, notif_type=NotificationType.STRUCTURE_LOST_ARMOR
            )


class TestNotificationRelevantWebhooks(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()

    def test_should_return_owner_webhooks_for_non_structure_notif(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.CHAR_APP_ACCEPT_MSG]
        )
        owner = OwnerFactory(webhooks=[webhook])
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.CHAR_APP_ACCEPT_MSG
        )
        # when
        result_qs = notif.relevant_webhooks()
        # then
        self.assertQuerysetEqual(result_qs, Webhook.objects.filter(pk=webhook.pk))

    def test_should_return_no_webhooks(self):
        # given
        owner = OwnerFactory(webhooks=False)
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.CHAR_APP_ACCEPT_MSG
        )
        # when
        result_qs = notif.relevant_webhooks()
        # then
        self.assertQuerysetEqual(result_qs, Webhook.objects.none())

    def test_should_return_owner_webhooks_for_structure_notif(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_UNDER_ATTACK]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = StructureFactory(owner=owner)
        notif = NotificationFactory(
            owner=owner,
            notif_type=NotificationType.STRUCTURE_UNDER_ATTACK,
            text_from_dict={
                "allianceID": 3011,
                "allianceLinkData": ["showinfo", 16159, 3011],
                "allianceName": "Big Bad Alliance",
                "armorPercentage": 98.65129050962584,
                "charID": 1011,
                "corpLinkData": ["showinfo", 2, 2011],
                "corpName": "Bad Company",
                "hullPercentage": 100.0,
                "shieldPercentage": 4.704536686417284e-14,
                "solarsystemID": structure.eve_solar_system_id,
                "structureID": structure.id,
                "structureShowInfoData": [
                    "showinfo",
                    structure.eve_type_id,
                    structure.id,
                ],
                "structureTypeID": structure.eve_type_id,
            },
        )
        # when
        result_qs = notif.relevant_webhooks()
        # then
        self.assertQuerysetEqual(result_qs, Webhook.objects.filter(pk=webhook.pk))

    def test_should_return_structure_webhooks_for_structure_notif(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_UNDER_ATTACK]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = StructureFactory(owner=owner)
        webhook_structure = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_UNDER_ATTACK]
        )
        structure.webhooks.add(webhook_structure)
        notif = NotificationFactory(
            owner=owner,
            notif_type=NotificationType.STRUCTURE_UNDER_ATTACK,
            text_from_dict={
                "allianceID": 3011,
                "allianceLinkData": ["showinfo", 16159, 3011],
                "allianceName": "Big Bad Alliance",
                "armorPercentage": 98.65129050962584,
                "charID": 1011,
                "corpLinkData": ["showinfo", 2, 2011],
                "corpName": "Bad Company",
                "hullPercentage": 100.0,
                "shieldPercentage": 4.704536686417284e-14,
                "solarsystemID": structure.eve_solar_system_id,
                "structureID": structure.id,
                "structureShowInfoData": [
                    "showinfo",
                    structure.eve_type_id,
                    structure.id,
                ],
                "structureTypeID": structure.eve_type_id,
            },
        )
        # when
        result_qs = notif.relevant_webhooks()
        # then
        self.assertQuerysetEqual(
            result_qs, Webhook.objects.filter(pk=webhook_structure.pk)
        )

    def test_should_return_owner_webhooks_when_notif_has_multiple_structures(self):
        # given
        webhook_owner = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REINFORCEMENT_CHANGED]
        )
        owner = OwnerFactory(webhooks=[webhook_owner])
        structure_1 = StructureFactory(owner=owner)
        webhook_structure = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REINFORCEMENT_CHANGED]
        )
        structure_1.webhooks.add(webhook_structure)
        structure_2 = StructureFactory(owner=owner)
        webhook_structure = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REINFORCEMENT_CHANGED]
        )
        structure_2.webhooks.add(webhook_structure)
        notif = NotificationFactory(
            owner=owner,
            notif_type=NotificationType.STRUCTURE_REINFORCEMENT_CHANGED,
            text_from_dict={
                "allStructureInfo": [
                    [
                        structure_1.id,
                        structure_1.name,
                        structure_1.eve_type_id,
                    ],
                    [
                        structure_2.id,
                        structure_2.name,
                        structure_2.eve_type_id,
                    ],
                ],
                "hour": 19,
                "numStructures": 2,
                "timestamp": 132141703753688216,
                "weekday": 255,
            },
        )
        # when
        result_qs = notif.relevant_webhooks()
        # then
        self.assertQuerysetEqual(result_qs, Webhook.objects.filter(pk=webhook_owner.pk))


class TestNotificationSendToConfiguredWebhooks(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_send_to_webhook(self, mock_send_to_webhook):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA]
        )
        owner = OwnerFactory(webhooks=[webhook])
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_configured_webhooks()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_send_to_webhook.called)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_send_to_multiple_webhooks(self, mock_send_to_webhook):
        # given
        webhook_1 = WebhookFactory(
            notification_types=[
                NotificationType.STRUCTURE_REFUELED_EXTRA,
                NotificationType.STRUCTURE_ANCHORING,
            ]
        )
        webhook_2 = WebhookFactory(
            notification_types=[
                NotificationType.STRUCTURE_REFUELED_EXTRA,
                NotificationType.STRUCTURE_DESTROYED,
            ]
        )
        owner = OwnerFactory(webhooks=[webhook_1, webhook_2])
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_configured_webhooks()
        # then
        self.assertTrue(result)
        self.assertEqual(mock_send_to_webhook.call_count, 2)
        webhook_pks = {call[0][0].pk for call in mock_send_to_webhook.call_args_list}
        self.assertSetEqual(webhook_pks, {webhook_1.pk, webhook_2.pk})

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_not_send_when_webhooks_are_inactive(self, mock_send_to_webhook):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
            is_active=False,
        )
        owner = OwnerFactory(webhooks=[webhook])
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_configured_webhooks()
        # then
        self.assertIsNone(result)
        self.assertFalse(mock_send_to_webhook.called)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_not_send_when_notif_types_dont_match(self, mock_send_to_webhook):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_UNDER_ATTACK]
        )
        owner = OwnerFactory(webhooks=[webhook])
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_configured_webhooks()
        # then
        self.assertIsNone(result)
        self.assertFalse(mock_send_to_webhook.called)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_return_false_when_sending_failed(self, mock_send_to_webhook):
        # given
        mock_send_to_webhook.return_value = False
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA]
        )
        owner = OwnerFactory(webhooks=[webhook])
        notif = NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_configured_webhooks()
        # then
        self.assertFalse(result)
        self.assertTrue(mock_send_to_webhook.called)

    @patch(MODULE_PATH + ".Notification.send_to_webhook")
    def test_should_send_to_structure_webhook(self, mock_send_to_webhook):
        # given
        webhook_owner = WebhookFactory(notification_types=[])
        owner = OwnerFactory(webhooks=[webhook_owner])
        structure = StructureFactory(owner=owner)
        webhook_structure = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_UNDER_ATTACK]
        )
        structure.webhooks.add(webhook_structure)
        notif = NotificationFactory(
            owner=owner,
            notif_type=NotificationType.STRUCTURE_UNDER_ATTACK,
            text_from_dict={
                "allianceID": 3011,
                "allianceLinkData": ["showinfo", 16159, 3011],
                "allianceName": "Big Bad Alliance",
                "armorPercentage": 98.65129050962584,
                "charID": 1011,
                "corpLinkData": ["showinfo", 2, 2011],
                "corpName": "Bad Company",
                "hullPercentage": 100.0,
                "shieldPercentage": 4.704536686417284e-14,
                "solarsystemID": structure.eve_solar_system_id,
                "structureID": structure.id,
                "structureShowInfoData": [
                    "showinfo",
                    structure.eve_type_id,
                    structure.id,
                ],
                "structureTypeID": structure.eve_type_id,
            },
        )
        # when
        result = notif.send_to_configured_webhooks()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_send_to_webhook.called)


@patch(MODULE_PATH + ".Webhook.send_message")
class TestNotificationSendToWebhook(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        EveEntityCorporationFactory(id=1000137, name="DED")

    def test_should_override_ping_type(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = StructureFactory(owner=owner)
        notif = Notification.create_from_structure(
            structure, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        notif.send_to_configured_webhooks(ping_type_override=Webhook.PingType.HERE)
        # then
        self.assertTrue(mock_send_message.called)
        _, kwargs = mock_send_message.call_args
        self.assertIn("@here", kwargs["content"])

    def test_should_override_color(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA]
        )
        owner = OwnerFactory(webhooks=[webhook])
        structure = StructureFactory(owner=owner)
        notif = Notification.create_from_structure(
            structure, notif_type=NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        notif.send_to_configured_webhooks(
            use_color_override=True, color_override=Webhook.Color.DANGER
        )
        # then
        self.assertTrue(mock_send_message.called)
        _, kwargs = mock_send_message.call_args
        self.assertEqual(kwargs["embeds"][0].color, Webhook.Color.DANGER)


@patch(MODULE_PATH + ".Webhook.send_message", spec=True)
class TestNotificationSendMessage(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):  # Can not be setUpTestData due to conflict with redis client
        super().setUpClass()
        load_eveuniverse()
        load_eve_entities()
        cls.owner = OwnerFactory(is_alliance_main=True)
        cls.webhook = cls.owner.webhooks.first()
        load_notification_entities(cls.owner)

    def test_can_send_message_normal(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        obj = clone_notification(Notification.objects.get(notification_id=1000020601))
        # when
        result = obj.send_to_webhook(self.webhook)
        # then
        self.assertTrue(result)
        _, kwargs = mock_send_message.call_args
        self.assertIsNotNone(kwargs["content"])
        self.assertIsNotNone(kwargs["embeds"])

    def test_should_mark_notification_as_sent_when_successful(self, mock_send_message):
        # given
        mock_send_message.return_value = True
        obj = NotificationFactory(owner=self.owner)
        # when
        obj.send_to_webhook(self.webhook)
        # then
        obj.refresh_from_db()
        self.assertTrue(obj.is_sent)

    def test_should_not_mark_notification_as_sent_when_error(self, mock_send_message):
        # given
        mock_send_message.return_value = 0
        obj = NotificationFactory(owner=self.owner)
        # when
        obj.send_to_webhook(self.webhook)
        # then
        obj.refresh_from_db()
        self.assertFalse(obj.is_sent)

    def test_can_send_all_notification_types(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        types_tested = set()
        # when /then
        for notif in Notification.objects.all():
            obj = clone_notification(notif)
            with self.subTest(notif_type=obj.notif_type):
                if obj.notif_type in NotificationType.values:
                    self.assertFalse(obj.is_sent)
                    self.assertTrue(obj.send_to_webhook(self.webhook))
                    types_tested.add(obj.notif_type)

        # make sure we have tested all existing esi notification types
        self.assertSetEqual(NotificationType.esi_notifications(), types_tested)

    def test_should_create_notification_for_structure_refueled(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        structure = StructureFactory(owner=self.owner, id=1000000000001)
        notif = Notification.create_from_structure(
            structure, NotificationType.STRUCTURE_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_webhook(self.webhook)
        # then
        self.assertTrue(result)

    def test_should_create_notification_for_tower_refueled(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        structure = StarbaseFactory(owner=self.owner, id=1300000000001)
        notif = Notification.create_from_structure(
            structure, NotificationType.TOWER_REFUELED_EXTRA
        )
        # when
        result = notif.send_to_webhook(self.webhook)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".STRUCTURES_DEFAULT_LANGUAGE", "en")
    def test_send_notification_without_existing_structure(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        obj = clone_notification(Notification.objects.get(notification_id=1000000505))
        # when
        obj.send_to_webhook(self.webhook)
        # then
        embed = mock_send_message.call_args[1]["embeds"][0]
        self.assertEqual(
            embed.description[:39], "The Astrahus **(unknown)** in [Amamake]"
        )

    @patch(MODULE_PATH + ".STRUCTURES_DEFAULT_LANGUAGE", "en")
    def test_notification_with_null_aggressor_alliance(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        obj = clone_notification(Notification.objects.get(notification_id=1000020601))
        # when
        result = obj.send_to_webhook(self.webhook)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SET_AVATAR", True)
    def test_can_send_message_with_setting_avatar(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        obj = clone_notification(Notification.objects.get(notification_id=1000020601))
        # when
        result = obj.send_to_webhook(self.webhook)
        # then
        self.assertTrue(result)
        _, kwargs = mock_send_message.call_args
        self.assertIsNotNone(kwargs["avatar_url"])
        self.assertIsNotNone(kwargs["username"])

    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SET_AVATAR", False)
    def test_can_send_message_without_setting_avatar(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        obj = clone_notification(Notification.objects.get(notification_id=1000020601))
        # when
        result = obj.send_to_webhook(self.webhook)
        # then
        self.assertTrue(result)
        _, kwargs = mock_send_message.call_args
        self.assertIsNone(kwargs["avatar_url"])
        self.assertIsNone(kwargs["username"])


@patch(MODULE_PATH + ".Webhook.send_message", spec=True)
class TestNotificationPings(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        corporation = EveEntityCorporationFactory()
        cls.notif_params = {
            "notif_type": NotificationType.BILLING_I_HUB_BILL_ABOUT_TO_EXPIRE,
            "text_from_dict": {
                "billID": 24803231,
                "corpID": corporation.id,
                "dueDate": 132936111600000000,
                "solarSystemID": 30000474,
            },
        }

    def test_can_ping(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory()
        owner = OwnerFactory(webhooks=[webhook])
        obj = NotificationFactory(owner=owner, **self.notif_params)
        # when
        result = obj.send_to_webhook(webhook)
        # then
        self.assertTrue(result)
        _, kwargs = mock_send_message.call_args
        self.assertIn("@everyone", kwargs["content"])

    def test_can_disable_pinging_webhook(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook_no_pings = WebhookFactory(has_default_pings_enabled=False)
        owner = OwnerFactory(webhooks=[webhook_no_pings])
        obj = NotificationFactory(owner=owner, **self.notif_params)
        # when
        result = obj.send_to_webhook(webhook_no_pings)
        self.assertTrue(result)
        _, kwargs = mock_send_message.call_args
        self.assertNotIn("@everyone", kwargs["content"])

    def test_can_disable_pinging_owner(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook_normal = WebhookFactory()
        owner = OwnerFactory(has_default_pings_enabled=False, webhooks=[webhook_normal])
        obj = NotificationFactory(owner=owner, **self.notif_params)
        # when
        result = obj.send_to_webhook(webhook_normal)
        # then
        self.assertTrue(result)
        _, kwargs = mock_send_message.call_args
        self.assertNotIn("@everyone", kwargs["content"])



from django.test import RequestFactory, TestCase

from structures.views import statistics

from ..testdata.factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
    OwnerFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    UserMainBasicFactory,
    UserMainDefaultFactory,
)
from ..testdata.load_eveuniverse import load_eveuniverse
from .utils import json_response_to_dict


class TestStatistics(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        alliance = EveAllianceInfoFactory(
            alliance_name="Wayne Enterprises", alliance_ticker="WYE"
        )
        cls.corporation = EveCorporationInfoFactory(
            corporation_name="Wayne Technologies", alliance=alliance
        )
        owner = OwnerFactory(corporation=cls.corporation)
        StructureFactory(owner=owner, eve_type_name="Astrahus")
        StructureFactory(owner=owner, eve_type_name="Athanor")
        PocoFactory.create_batch(size=4, owner=owner)
        StarbaseFactory.create_batch(size=3, owner=owner)
        cls.character = EveCharacterFactory(corporation=cls.corporation)

    def test_should_return_summary_data(self):
        # given
        user = UserMainDefaultFactory(main_character__character=self.character)
        # when
        request = self.factory.get("/")
        request.user = user
        response = statistics.structure_summary_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        obj = data[self.corporation.corporation_id]
        self.assertEqual(obj["corporation_name"], "Wayne Technologies")
        self.assertEqual(obj["alliance_name"], "Wayne Enterprises [WYE]")
        self.assertEqual(obj["citadel_count"], 1)
        self.assertEqual(obj["ec_count"], 0)
        self.assertEqual(obj["refinery_count"], 1)
        self.assertEqual(obj["other_count"], 0)
        self.assertEqual(obj["poco_count"], 4)
        self.assertEqual(obj["starbase_count"], 3)
        self.assertEqual(obj["total"], 9)

    def test_should_return_no_summary_data_without_permission(self):
        # given
        user = UserMainBasicFactory(main_character__character=self.character)
        # when
        request = self.factory.get("/")
        request.user = user
        response = statistics.structure_summary_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertFalse(data)



# flake8: noqa
"""
this scripts create a test owner and adds test notifications to it
"""

import datetime as dt
import json
from pathlib import Path

from app_utils.scripts import start_django

start_django()


def main():

    from django.utils.timezone import now
    from esi.clients import esi_client_factory
    from eveuniverse.models import EveEntity

    from allianceauth.eveonline.models import EveCorporationInfo

    from structures.core.notification_types import NotificationType
    from structures.models import Notification, Owner, Structure, Webhook

    # corporation / structure the notifications will be added to
    CORPORATION_ID = 1000127  # Guristas

    print(
        "load_test_notifications - "
        "script loads test notification into the local database "
    )

    print("Connecting to ESI ...")
    client = esi_client_factory()

    print("Creating base data ...")
    try:
        corporation = EveCorporationInfo.objects.get(corporation_id=CORPORATION_ID)
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(CORPORATION_ID)

    owner, created = Owner.objects.get_or_create(
        corporation=corporation, defaults={"is_active": False}
    )
    if created and not owner.webhooks.exists():
        webhook = Webhook.objects.filter(is_active=True, is_default=True).first()
        if webhook:
            owner.webhooks.add(webhook)

    structure = {
        "fuel_expires": None,
        "name": "Test Structure Alpha",
        "next_reinforce_apply": None,
        "next_reinforce_hour": None,
        "position": {"x": 55028384780.0, "y": 7310316270.0, "z": -163686684205.0},
        "profile_id": 101853,
        "reinforce_hour": 18,
        "services": [
            {
                "name": "Clone Bay",
                "name_de": "Clone Bay_de",
                "name_ko": "Clone Bay_ko",
                "state": "online",
            },
            {
                "name": "Market Hub",
                "name_de": "Market Hub_de",
                "name_ko": "Market Hub_ko",
                "state": "offline",
            },
        ],
        "state": "shield_vulnerable",
        "state_timer_end": None,
        "state_timer_start": None,
        "structure_id": 1999999999999,
        "system_id": 30002537,
        "type_id": 35832,
        "unanchors_at": None,
    }
    structure, _ = Structure.objects.update_or_create_from_dict(structure, owner)

    p = Path(__file__).parent / "/entities.json"
    with p.open(mode="r", encoding="utf-8") as fp:
        data = json.load(fp)

    notifications = data["Notification"]
    for n in notifications:
        if n["sender_id"] == 2901:
            n["sender_id"] = 1000137  # DED
        if n["sender_id"] == 2902:
            n["sender_id"] = 1000125  # Concord
        elif n["sender_id"] == 1011:
            n["sender_id"] = 3004029
        elif n["sender_id"] == 2022:
            n["sender_id"] = 1000127  # Guristas
        elif n["sender_id"] == 3001:
            n["sender_id"] = 99010298
        n["text"] = n["text"].replace("1000000000001", str(structure.id))
        n["text"] = n["text"].replace("35835", str(structure.eve_type_id))
        n["text"] = n["text"].replace("35835", str(structure.eve_type_id))
        n["text"] = n["text"].replace("30002537", str(structure.eve_solar_system_id))
        n["text"] = n["text"].replace("1001", "3004037")
        n["text"] = n["text"].replace("1002", "3019491")
        n["text"] = n["text"].replace("1011", "3004029")
        n["text"] = n["text"].replace("2001", "98394960")
        n["text"] = n["text"].replace("2002", "1000134")  # Blood Raiders
        n["text"] = n["text"].replace("3001", "99005502")
        n["text"] = n["text"].replace("3002", "99009333")
        n["text"] = n["text"].replace("3011", "1354830081")

    timestamp_start = now() - dt.timedelta(hours=2)

    for n in notifications:
        notif_type = n["type"]
        if notif_type not in NotificationType.values:
            print(f"Skipping unsupported: {notif_id} {notif_type}")

        notif_id = n["notification_id"]
        sender, _ = EveEntity.objects.get_or_create_esi(id=n["sender_id"])
        text = n["text"] if "text" in n else None
        is_read = n["is_read"] if "is_read" in n else None
        timestamp_start = timestamp_start + dt.timedelta(minutes=5)
        obj, created = Notification.objects.update_or_create(
            notification_id=notif_id,
            owner=owner,
            defaults={
                "sender": sender,
                "timestamp": timestamp_start,
                "notif_type": notif_type,
                "text": text,
                "is_read": is_read,
                "last_updated": now(),
                "is_sent": False,
            },
        )
        if created:
            print(f"Created: {notif_id} {notif_type}")

    print("DONE")

    """
    for notification in notifications:
        dt = datetime.datetime.utcfromtimestamp(notification['timestamp'])
        dt = pytz.utc.localize(dt)
        notification['timestamp'] = dt.isoformat()

    with open(
        file=currentdir + '/td_notifications_2.json',
        mode='w',
        encoding='utf-8'
    ) as f:
        json.dump(
            notifications,
            f,
            sort_keys=True,
            indent=4
        )

    """


if __name__ == "__main__":
    main()



from .eveuniverse import EveSovereigntyMap, EveSpaceType
from .notifications import (
    FuelAlert,
    FuelAlertConfig,
    GeneratedNotification,
    JumpFuelAlert,
    JumpFuelAlertConfig,
    Notification,
    Webhook,
    get_default_notification_types,
)
from .owners import Owner, OwnerCharacter
from .structures_1 import Structure, StructureItem, StructureTag
from .structures_2 import (
    PocoDetails,
    StarbaseDetail,
    StarbaseDetailFuel,
    StructureService,
)

__all__ = [
    "Owner",
    "OwnerCharacter",
    "EveSovereigntyMap",
    "EveSpaceType",
    "PocoDetails",
    "StarbaseDetail",
    "StarbaseDetailFuel",
    "Structure",
    "StructureItem",
    "StructureService",
    "StructureTag",
    "FuelAlert",
    "FuelAlertConfig",
    "GeneratedNotification",
    "JumpFuelAlert",
    "JumpFuelAlertConfig",
    "Notification",
    "Webhook",
    "get_default_notification_types",
]



# this scripts start a large number of tasks for load testing

import inspect
import os
import sys

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(currentdir))))
    + "/myauth"
)
sys.path.insert(0, myauth_dir)

import django  # noqa: E402
from django.apps import apps  # noqa: E402

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

from celery import chain  # noqa: E402

from structures.tasks import fetch_all_notifications  # noqa: E402

if not apps.is_installed("structures"):
    raise RuntimeError("The app structures is not installed")


NUMBER_OF_TASKS = 100

print("Starting %d tasks..." % NUMBER_OF_TASKS)
task_list = list()
for _ in range(NUMBER_OF_TASKS):
    task_list.append(fetch_all_notifications.si())

chain(task_list).delay()
print("Tasks started")



"""Views for public page."""

from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from structures import __title__
from structures.constants import EveCategoryId
from structures.core.serializers import PocoListSerializer
from structures.models import Structure

from .common import add_common_context, add_common_data_export

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("structures.basic_access")
def public(request: HttpRequest) -> HttpResponse:
    """Return view to render Public page."""
    characters = (
        EveCharacter.objects.filter(character_ownership__user=request.user)
        .order_by("character_name")
        .values("character_id", "character_name")
    )
    character_id = int(request.GET.get("character_id", 0))
    if not character_id:
        try:
            character_id = request.user.profile.main_character.character_id
        except AttributeError:
            character_id = None

    pocos_count = _public_pocos_query().count()
    selected_character = get_object_or_404(EveCharacter, character_id=character_id)

    ajax_url = reverse(
        "structures:public_poco_list_data", args=[selected_character.character_id]
    )
    data_export = add_common_data_export(
        {
            "ajax_url": ajax_url,
            "filter_titles": {
                "alliance": _("Alliance"),
                "access": _("Access?"),
                "corporation": _("Corporation"),
                "constellation": _("Constellation"),
                "planet_type": _("Planet Type"),
                "region": _("Region"),
                "space_type": _("Space Type"),
                "solar_system": _("Solar System"),
            },
        }
    )
    context = {
        "characters": characters,
        "selected_character": selected_character,
        "pocos_count": pocos_count,
        "data_export": data_export,
    }
    return render(request, "structures/public.html", add_common_context(context))


def public_poco_list_data(request: HttpRequest, character_id: int) -> JsonResponse:
    """List of public POCOs for DataTables."""
    character = get_object_or_404(EveCharacter, character_id=character_id)
    pocos = _public_pocos_query()
    serializer = PocoListSerializer(
        queryset=pocos, request=request, character=character
    )
    data = serializer.to_list()
    return JsonResponse({"data": data})


def _public_pocos_query():
    pocos = Structure.objects.select_related_defaults().filter(
        eve_type__eve_group__eve_category_id=EveCategoryId.ORBITAL,
        owner__are_pocos_public=True,
    )
    return pocos



"""Eve sovereignty related core logic."""

from typing import Optional

from structures.constants import EveTypeId

_MAP_CAMPAIGN_EVENT_2_TYPE_ID = {
    1: EveTypeId.TCU,
    2: EveTypeId.IHUB,
}
_MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME = {
    EveTypeId.CUSTOMS_OFFICE: "POCO",
    EveTypeId.TCU: "TCU",
    EveTypeId.IHUB: "I-HUB",
}


def event_type_to_type_id(event_type: int) -> Optional[int]:
    """Convert an event type to a type ID."""
    return _MAP_CAMPAIGN_EVENT_2_TYPE_ID.get(event_type)


def event_type_to_structure_type_name(event_type: int) -> str:
    """Convert an event type to a structure type name."""
    return _MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME.get(
        event_type_to_type_id(event_type), "Other"
    )
